#include <stdio.h>
#include <math.h>
#include "opcode.h"

//typedef unsigned int uint;

union REG 
{
	uint u;
	int i;
	float f;
};

uint xor(uint a, uint b){
  return ~a & b | a & ~b;
}


uint pc;
uint inst;
union REG reg[32];

#define INST_MEM_SIZE 32
#define DATA_MEM_SIZE 4096

unsigned int inst_mem[INST_MEM_SIZE];
unsigned int data_mem[DATA_MEM_SIZE];

void calc();
void show();

int main(int argc, char *argv[])
{
	FILE *fp;
	int i;
	
	if(argc == 2)
		fp = fopen(argv[1], "r");
	else
		fp = NULL;
	
	if(!fp)
	{
		printf("usage: %s input_file\n",argv[0]);
		return(1);
	}
	
	for(i=0;!feof(fp);i++)
		fscanf(fp, "%*c%*c%08X\n", inst_mem + i);
	
	pc = 0;
	reg[0].u = 0;
	
	calc();
	
	show();
	
	return 0;
}

void show()
{
	int i;
	
	printf("PC : 0x%08X\n", pc);
	printf("INST : 0x%08X\n", inst);
	
	for(i=0;i<16;i++)
		printf("r%2d : 0x%08X\tr%2d : 0x%08X\n", i, reg[i].u, i+16, reg[i+16].u);
	for(i=0;i<16;i++)
	  printf("m%2d : 0x%08X\tm%2d : 0x%08X\n", i, data_mem[DATA_MEM_SIZE - i - 1], i + 16, data_mem[DATA_MEM_SIZE - i - 17]);

}

void calc()
{
	printf("0x%08X\n", pc);
	inst = inst_mem[pc];
	show();
	
	uint op = inst >> 26;
	uint rs = inst >> 21 & 0x1F;
	uint rt = inst >> 16 & 0x1F;
	uint rd = inst >> 11 & 0x1F;
	uint shft = inst >> 6 & 0x1F;
	uint fnct = inst & 0x3F;
	int imm = (inst & 0x8000) ? inst & 0xFFFF ^ 0xFFFF0000 : inst & 0xFFFF;
	int target = ((inst & 0x2000000) ? 0xFC << 24 : 0) | inst & 0x3FFFFFF;
	
	switch(op)
	{
	/**********************************/
	/* ALU OPERATIONS                 */
	/**********************************/
	case OP_ALU:
		switch(fnct)
		{
		case FNCT_ADD:
			reg[rd].i = reg[rs].i + reg[rt].i;
			break;
		case FNCT_SUB:
			reg[rd].i = reg[rs].i - reg[rt].i;
			break;
		case FNCT_MUL:
			reg[rd].i = reg[rs].i * reg[rt].i;
			break;
		case FNCT_DIV:
			reg[rd].i = reg[rs].i / reg[rt].i;
			break;
		case FNCT_SLL:
			reg[rd].u = reg[rs].u << shft;
			break;
		case FNCT_SRL:
			reg[rd].u = reg[rs].u >> shft;
			break;
		case FNCT_SRR:
			reg[rd].i = reg[rs].i >> shft;
			break;
		case FNCT_AND:
			reg[rd].u = reg[rs].u & reg[rt].u;
			break;
		case FNCT_OR:
			reg[rd].u = reg[rs].u | reg[rt].u;
			break;
		case FNCT_XOR:
			reg[rd].u = xor(reg[rs].u, reg[rt].u);
			break;
		case FNCT_NOT:
			reg[rd].u = ~reg[rs].u;
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* IMMEDIATE OPERATIONS           */
	/**********************************/
	case OP_ADDI:
		reg[rt].i = reg[rs].i + imm;
		pc++;
		break;
	case OP_SUBI:
		reg[rt].i = reg[rs].i - imm;
		pc++;
		break;
	case OP_MULI:
		reg[rt].i = reg[rs].i * imm;
		pc++;
		break;
	case OP_DIVI:
		reg[rt].i = reg[rs].i / imm;
		pc++;
		break;
	case OP_ANDI:
		reg[rt].u = reg[rs].u & imm;
		pc++;
		break;
	case OP_ORI:
		reg[rt].u = reg[rs].u | imm;
		pc++;
		break;
	case OP_XORI:
		reg[rt].u = xor(reg[rs].u, (uint)imm);
		pc++;
		break;
	
	/**********************************/
	/* FPU OPERATIONS                 */
	/**********************************/
	case OP_FPU:
		switch(fnct)
		{
		case FNCT_FADD:
			reg[rd].f = reg[rs].f + reg[rt].f;
			break;
		case FNCT_FSUB:
			reg[rd].f = reg[rs].f - reg[rt].f;
			break;
		case FNCT_FMUL:
			reg[rd].f = reg[rs].f * reg[rt].f;
			break;
		case FNCT_FDIV:
			reg[rd].f = reg[rs].f / reg[rt].f;
			break;
		case FNCT_ITOF:
			reg[rd].f = (float)reg[rs].i;
			break;
		case FNCT_FTOI:
			reg[rd].i = (int)reg[rs].f;
			break;
		case FNCT_FABS:
			reg[rd].f = fabsf(reg[rs].f);
			break;
		case FNCT_FNEG:
			reg[rd].f = -reg[rs].f;
			break;
		case FNCT_FLOOR:
			reg[rd].f = floorf(reg[rs].f);
			break;
		case FNCT_FSQRT:
			reg[rd].f = (float)sqrt(reg[rs].f);	//sqrtf
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* SYSTEM OPERATIONS              */
	/**********************************/
	case OP_SYS:
		switch(fnct){
		case FNCT_NOP:
			break;
		case FNCT_HLT:
			printf("system halt...\n");
			return;
		case FNCT_RD:
			scanf("%08X[ \t\n]", &reg[rd].u);
			break;
		case FNCT_WR:
			printf("%08X\n", reg[rs].u);
			break;
		}
		pc++;
		break;
	
	/**********************************/
	/* MEMORY OPERATIONS              */
	/**********************************/
	case OP_LOAD:
	  reg[rt].u = data_mem[(reg[rs].u + imm) % DATA_MEM_SIZE];
		pc++;
		break;
	case OP_STORE:
	  data_mem[(reg[rs].u + imm) % DATA_MEM_SIZE] = reg[rt].u;
		pc++;
		break;
	
	/**********************************/
	/* JUMP OPERATIONS                */
	/**********************************/
	case OP_JUMP:
		pc = pc + target;
		break;
	case OP_JAL:
		reg[31].u = pc + 1;		//1 or 4?
		pc = pc + (unsigned int)target;
		break;
	case OP_JR:
		pc = reg[rs].u;
		break;
	
	/**********************************/
	/* INT BRANCH OPERATIONS          */
	/**********************************/
	case OP_BL  :
		if(reg[rs].i <  reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BE  :                  
		if(reg[rs].i == reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BLE :                  
		if(reg[rs].i <= reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BGE :                  
		if(reg[rs].i >= reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BNE :                  
		if(reg[rs].i != reg[rt].i) pc += imm; else pc ++;
		break;                 
	case OP_BG  :                  
		if(reg[rs].i >  reg[rt].i) pc += imm; else pc ++;
		break;
	case OP_BLZ :
		if(reg[rs].i <  0) pc += imm; else pc ++;
		break;    
	case OP_BEZ :     
		if(reg[rs].i == 0) pc += imm; else pc ++;
		break;    
	case OP_BLEZ:     
		if(reg[rs].i <= 0) pc += imm; else pc ++;
		break;    
	case OP_BGEZ:     
		if(reg[rs].i >= 0) pc += imm; else pc ++;
		break;    
	case OP_BNEZ:     
		if(reg[rs].i != 0) pc += imm; else pc ++;
		break;    
	case OP_BGZ :     
		if(reg[rs].i >  0) pc += imm; else pc ++;
		break;
	/**********************************/
	/* FLOAT BRANCH OPERATIONS        */
	/**********************************/
	case OP_FBL  :
		if(reg[rs].f <  reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBE  :                 
		if(reg[rs].f == reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBLE :                 
		if(reg[rs].f <= reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBGE :                 
		if(reg[rs].f >= reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBNE :                 
		if(reg[rs].f != reg[rt].f) pc += imm; else pc ++;
		break;                 
	case OP_FBG  :                 
		if(reg[rs].f >  reg[rt].f) pc += imm; else pc ++;
		break;
	case OP_FBLZ :
		if(reg[rs].f <  0) pc += imm; else pc ++;
		break;    
	case OP_FBEZ :     
		if(reg[rs].f == 0) pc += imm; else pc ++;
		break;    
	case OP_FBLEZ:     
		if(reg[rs].f <= 0) pc += imm; else pc ++;
		break;    
	case OP_FBGEZ:     
		if(reg[rs].f >= 0) pc += imm; else pc ++;
		break;    
	case OP_FBNEZ:     
		if(reg[rs].f != 0) pc += imm; else pc ++;
		break;    
	case OP_FBGZ :    
		if(reg[rs].f >  0) pc += imm; else pc ++;
		break;
	}
	
	printf("press <ENTER> to continue...\n");
	getchar();
	
	calc();
}


